/**************************************
 * $Author: kaneda $
 * $Date: 2003/03/12 09:04:36 $
 **************************************/

/* addr.c */
#include "ph_addr.h"


/* Public Function Prototypes */
static char * get_host(char * addr);
static int is_localhost(char * host);
static void ph_addr_print(ph_addr_t addr, void * arg);


ph_addr_t ph_addr_new(char * host, char * service, ph_addr_type_t type, char * user) {
  ph_addr_t addr;

  assert(host != NULL);
  assert(service != NULL);

  addr = (ph_addr_t)ph_malloc(sizeof(struct ph_addr));
  addr->host = ph_strdup(host);
  addr->service = ph_strdup(service);
  addr->type = type;
  addr->user = (user != NULL) ? ph_strdup(user) : NULL;

  return addr;
}


void ph_addr_dispose(ph_addr_t addr) { 
  if (addr == NULL) {
    ph_err_msg("ph_addr_dispose() : NULL input");
    return;
  }

  if (addr->user != NULL) {
    ph_free(addr->user);
  }

  ph_free(addr->host);
  ph_free(addr->service);
  ph_free(addr);
}


ph_addr_t ph_addr_dup(ph_addr_t addr) {
 return ph_addr_new(addr->host, addr->service, addr->type, addr->user);
}


static char * get_host(char * addr) {
  struct hostent * hp;    
  struct sockaddr_in sin;
  
  assert(addr != NULL);

  /* TODO : MT-unsafe */
  if ((hp = gethostbyname(addr)) == NULL) {
    //ph_err_msg("gethostbyname(%s) returns NULL", addr);
    return NULL;
  } 
 
  memcpy(&sin.sin_addr.s_addr, *hp->h_addr_list, sizeof(sin.sin_addr.s_addr));
  return ph_sock_ntop((struct sockaddr *)&sin, sizeof(sin)); 
}


static int is_localhost(char * host) {
  char s[PH_BUFSIZE];

  if (ph_streq(host, "127.0.0.1")) {
    return 1;
  }

  ph_mzero(s, 8192);
  if (gethostname(s, 8192) == -1) {
    return 0;
  } 


  return ph_streq(host, get_host(s));
}


char * ph_addr_get_host(char * addr, int service, int listen_port) {
  char * host;
  char * s;


  if ((s = get_host(addr)) == NULL) {
    return NULL;
  }
  
  host = ph_strdup(s);

  if ((is_localhost(host)) && (listen_port == service)) { /* omit a connection to localhost */
    ph_free(host);
    return NULL;
  }
  
  return host;
}


PH_DEFINE_LIST_FUNC_BODY(ph_addr) 


static void ph_addr_print(ph_addr_t addr, void * arg) {
  ph_debug_print("(%p) type=%s, host=%s, service=%s\n",
		 addr,
		 ((addr->type == tcp) ? "tcp" : "ssh"),
		 ((addr->host != NULL) ? addr->host: "(null)"),
		 ((addr->service != NULL) ? addr->service: "(null)"));
}


void ph_addr_list_print(ph_addr_list_t addr_list) {
  assert(addr_list != NULL);

  if (ph_addr_list_is_empty(addr_list)) {
    ph_debug_print2("[addr_list]\n");  
  } else {
    ph_debug_print2("[addr_list \n");  
    ph_addr_list_map(addr_list, &ph_addr_print, NULL);
    ph_debug_print2("]\n");
  }
}


/* TODO : ph_conn_list_init_loop $B$H$&$^$/E}0l$G$-$=$&$@$C$?$i!$E}0l$9$k(B*/
/* for synchronized iteration */
/* Note : ph_conn_list_init_loop manages dynamic addition/removal of cells. ph_addr_list_init_loop manages only dynamic addition of cells. */
ph_addr_cell_t ph_addr_list_init_loop(ph_addr_list_t addr_list) {
  ph_addr_cell_t p;

  assert(addr_list != NULL);

  ph_pthread_mutex_lock(&addr_list->mp);
  p = addr_list->head;
  ph_pthread_mutex_unlock(&addr_list->mp);  

  return p;
}


ph_addr_cell_t ph_addr_list_iterate(ph_addr_list_t addr_list, ph_addr_cell_t p) {
  ph_addr_cell_t next;
  
  assert(addr_list != NULL);
  assert(p != NULL);

  ph_pthread_mutex_lock(&addr_list->mp);
  next = p->next;
  ph_pthread_mutex_unlock(&addr_list->mp);

  return next;
}
