/**************************************
 * $Author: kaneda $
 * $Date: 2003/02/12 12:45:20 $
 **************************************/

/* conn.c */
#include "ph_conn.h"


/* Public Function Prototypes */
static void ph_conn_print(ph_conn_t conn, void * arg);
static int is_successful_connect();

ph_conn_t ph_conn_new() {
  ph_conn_t conn;

  conn = (ph_conn_t)ph_malloc(sizeof(struct ph_conn));
  conn->addr = NULL;
  conn->state = closed;
  conn->attr = NULL;
  conn->connect_func_p = NULL;
  conn->send_func_p = NULL;
  conn->recv_func_p = NULL;
  conn->close_func_p = NULL;
  conn->dispose_func_p = NULL;
  conn->is_authed = 0;
  conn->is_repaired = 0;
#ifdef PH_ROUTING_ENABLE_DOMINATING
  conn->is_permanent = 0;
#endif
  conn->nref = 0;
  ph_pthread_mutex_init(&conn->mp, NULL);
  return conn;
}


void ph_conn_dispose(ph_conn_t conn) { 
  if (conn == NULL) {
    ph_err_msg("ph_conn_dispose() : NULL input");
    return; 
  }

  assert(conn->nref == 0);
  //assert(conn->state == closed);


  if (conn->addr != NULL) {
    ph_addr_dispose(conn->addr);
  }

  if (conn->attr != NULL) {
    ph_conn_close(conn);
    assert(conn->dispose_func_p != NULL);
    (*conn->dispose_func_p)(conn->attr);
  }

  ph_pthread_mutex_destroy(&conn->mp);
  ph_free(conn);
}


static void ph_conn_set_funcs(ph_conn_t conn) {
  assert(conn != NULL);
  assert(conn->addr != NULL);

  switch (conn->addr->type) {
  case tcp:
  case unknown:
    conn->connect_func_p = &ph_tcp_attr_connect;
    conn->listen_func_p = &ph_tcp_attr_listen;
    conn->accept_func_p = &ph_tcp_attr_accept;
    conn->send_func_p = &ph_tcp_attr_send;
    conn->recv_func_p = &ph_tcp_attr_recv;
    conn->close_func_p = &ph_tcp_attr_close;
    conn->dispose_func_p = &ph_tcp_attr_dispose;
    conn->fd_isset_func_p = &ph_tcp_attr_fd_isset;
    conn->fd_set_func_p = &ph_tcp_attr_fd_set;
    break;
  case ssh:
    conn->connect_func_p = &ph_ssh_attr_connect;
    conn->listen_func_p = &ph_tcp_attr_listen;
    conn->accept_func_p = &ph_tcp_attr_accept;
    conn->send_func_p = &ph_ssh_attr_send;
    conn->recv_func_p = &ph_ssh_attr_recv;
    conn->close_func_p = &ph_ssh_attr_close;
    conn->dispose_func_p = &ph_ssh_attr_dispose;
    conn->fd_isset_func_p = &ph_ssh_attr_fd_isset;
    conn->fd_set_func_p = &ph_ssh_attr_fd_set;
    break;
  default:
    ph_err_quit("ph_conn_set_funcs() : no matching");
    break;
  }  
}


ph_conn_t ph_conn_connect(ph_addr_t addr) {
  ph_conn_t conn;

  assert(addr != NULL);

  conn = ph_conn_new();
  conn->addr = ph_addr_dup(addr);
  ph_conn_set_funcs(conn);

  assert(conn->connect_func_p != NULL);
  if ((conn->attr = (*conn->connect_func_p)(conn->addr)) == NULL) {
    // connect fail
    ph_conn_dispose(conn);
    return NULL;
  }

  conn->state = opened;
  
  //conn->state = connecting; TODO : non-block
  ph_debug_print("ph_conn_connect()\t\t: \n");

  return conn;
}


static int is_successful_connect() {
  return 1; /* TODO: non-blocking connect() $B$,@.8y$7$?>l9g(B */
}


/* TODO */
int ph_conn_connecting(ph_conn_t conn) {
  if (is_successful_connect() == 0) {
    ph_conn_set_state(conn, closed);
    return 0;
  }
  ph_conn_set_state(conn, opened);
  return 1;
}

ph_conn_t ph_conn_listen(int available_ports[2], int * listen_port_p) {
  ph_conn_t conn;
  int port;

  ph_debug_print("ph_conn_listen()\t\t: \n");  
  conn = ph_conn_new(); 
  conn->state = accepting;

  for (port = available_ports[0]; port <= available_ports[1]; port++) {
    char service[PH_BUFSIZE];

    ph_mzero(service, PH_BUFSIZE);
    ph_snprintf(service, PH_BUFSIZE, "%d", port);   
    
    /* TODO : only IPv4 */
    conn->addr = ph_addr_new("0.0.0.0", service, tcp, NULL);
    ph_conn_set_funcs(conn);

    assert(conn->listen_func_p != NULL);
    if ((conn->attr = (*conn->listen_func_p)(conn->addr)) != NULL) {
      (*listen_port_p) = port;
      return conn;
    }
    ph_addr_dispose(conn->addr);
  }

  ph_err_quit("bind error");
  return NULL;
}


ph_conn_t ph_conn_accept(ph_conn_t listen_conn) {
  ph_conn_t conn;
  struct sockaddr cliaddr;
  socklen_t len = sizeof(cliaddr);

  ph_debug_print("ph_conn_accept()\t\t: \n");  

  assert(listen_conn != NULL);
  assert(listen_conn->accept_func_p != NULL);

  conn = ph_conn_new(); 
  if ((conn->attr = (*listen_conn->accept_func_p)(listen_conn->attr, &cliaddr, &len)) == NULL) {
    ph_conn_dispose(conn);
    return NULL;
  }

  //conn->addr = ph_addr_new(ph_sock_ntop(&cliaddr, len), "xxxx", tcp); /* TODO */
  conn->addr = ph_addr_new("xxxx", "xxxx", unknown, NULL); /* TODO */
  ph_conn_set_funcs(conn);
  conn->state = opened;  

  return conn;
}


ssize_t ph_conn_send(ph_conn_t conn, void * buf, size_t len) {
  ssize_t n;

  assert(conn != NULL);
  ph_debug_print("ph_conn_send()\t\t\t: \n");

  assert(conn->send_func_p != NULL);
  n = (*conn->send_func_p)(conn->attr, buf, len);

  ph_debug_print_message(conn->mp, buf, len);
  return n;
}


ssize_t ph_conn_recv(ph_conn_t conn, void * buf, size_t len) {
  ssize_t n;

  assert(conn != NULL);
  ph_debug_print("ph_conn_recv()\t\t\t: \n");

  assert(conn->recv_func_p != NULL);
  n = (*conn->recv_func_p)(conn->attr, buf, len);

  ph_debug_print_message(conn->mp, buf, len);
  return n;
}


void ph_conn_close(ph_conn_t conn) {
  assert(conn != NULL);
  assert(conn->close_func_p != NULL);

  ph_debug_print("ph_conn_close()\t\t\t: \n");
  
  (*conn->close_func_p)(conn->attr);
  conn->state = closed;
}


void ph_conn_lock(ph_conn_t conn) {
  assert(conn != NULL);

  ph_pthread_mutex_lock(&conn->mp);
  conn->nref++;
  ph_pthread_mutex_unlock(&conn->mp);
}


void ph_conn_unlock(ph_conn_t conn) {
  assert(conn != NULL);
  assert(conn->nref > 0);

  ph_pthread_mutex_lock(&conn->mp);
  conn->nref--;
  ph_pthread_mutex_unlock(&conn->mp);
}


ph_conn_state_t	ph_conn_get_state(ph_conn_t conn) {
  ph_conn_state_t state;

  assert(conn != NULL);
  ph_pthread_mutex_lock(&conn->mp);
  state = conn->state;
  ph_pthread_mutex_unlock(&conn->mp);  
  return state;
}


void ph_conn_set_state(ph_conn_t conn, ph_conn_state_t state) {
  assert(conn != NULL);
  ph_pthread_mutex_lock(&conn->mp);
  conn->state = state;
  ph_pthread_mutex_unlock(&conn->mp);
}


ph_addr_type_t ph_conn_get_addr_type(ph_conn_t conn) {
  ph_addr_type_t type;

  assert(conn != NULL);
  
  if (conn->addr == NULL) {
    ph_err_msg(" ph_conn_get_addr_type()\t: addr == NULL");
    return unknown;
  }

  ph_pthread_mutex_lock(&conn->mp);
  type = conn->addr->type;
  ph_pthread_mutex_unlock(&conn->mp);  
  return type;
}


void ph_conn_set_addr_type(ph_conn_t conn, ph_addr_type_t type) {
  assert(conn != NULL);
  assert(conn->addr != NULL);

  ph_pthread_mutex_lock(&conn->mp);
  conn->addr->type = type;
  ph_pthread_mutex_unlock(&conn->mp);  
}


#ifdef PH_ROUTING_ENABLE_DOMINATING
int ph_conn_get_is_permanent(ph_conn_t conn) {
  int is_permanent;

  assert(conn != NULL);
  ph_pthread_mutex_lock(&conn->mp);
  is_permanent = conn->is_permanent;
  ph_pthread_mutex_unlock(&conn->mp);  
  return is_permanent;
}


void ph_conn_set_is_permanent(ph_conn_t conn, int is_permanent) {
  assert(conn != NULL);
  ph_pthread_mutex_lock(&conn->mp);
  conn->is_permanent = is_permanent;
  ph_pthread_mutex_unlock(&conn->mp);  
}
#endif

PH_DEFINE_LIST_FUNC_BODY(ph_conn) 


static char * state_print(ph_conn_state_t state) {
  switch (state) {
  case connecting:
    return "connecting";
    break;
  case accepting:
    return "accepting";
    break;
  case opened:
    return "opened";
    break;
  case closed:
    return "closed";
    break;
  default:
    ph_err_quit("state_print() : no matching (state=%d)", state);
    break;
  }
  return NULL;
}

static void ph_conn_print(ph_conn_t conn, void * arg) {
  ph_debug_print("(%p) type=%s, host=%s, port=%s, state=%s, nref=%d\n",
		 conn, 
		 ((conn->addr != NULL) ? ((conn->addr->type == tcp) ? "tcp" : (conn->addr->type == ssh) ? "ssh" : "unknown") : "(null)"),
		 (((conn->addr != NULL) && (conn->addr->host != NULL)) ? conn->addr->host : "(null)"), 
		 (((conn->addr != NULL) && (conn->addr->service != NULL)) ? conn->addr->service : "(null)"), 

		 state_print(conn->state), 
		 conn->nref);
}

void ph_conn_list_print(ph_conn_list_t conn_list) {
  assert(conn_list != NULL);

  if (ph_conn_list_is_empty(conn_list)) {
    ph_debug_print2("[conn_list]\n");
  } else {
    ph_debug_print2("[conn_list \n");
    ph_conn_list_map(conn_list, &ph_conn_print, NULL);
    ph_debug_print2("]\n");
  }
}


static int is_same_addr(ph_conn_t conn, void * arg) {
  ph_addr_t addr = (ph_addr_t)arg;

  return ((conn->addr != NULL) && 
	  (ph_streq(conn->addr->host, "xxxx") == 0) && // not accepted connection
	  (ph_streq(conn->addr->host, addr->host)) &&
	  (ph_streq(conn->addr->service, addr->service)));
}

ph_conn_t ph_conn_list_find_same_addr(ph_conn_list_t conn_list, ph_addr_t addr) {
  return ph_conn_list_find(conn_list, &is_same_addr, (void *)addr);
}


/* for synchronized iteration */
ph_conn_cell_t ph_conn_list_init_loop(ph_conn_list_t conn_list) {
  ph_conn_cell_t p;

  assert(conn_list != NULL);

  ph_pthread_mutex_lock(&conn_list->mp);
  p = conn_list->head;
  if (p != NULL) {
    ph_conn_lock(p->self);
  }
  ph_pthread_mutex_unlock(&conn_list->mp);  

  return p;
}


ph_conn_cell_t ph_conn_list_iterate(ph_conn_list_t conn_list, ph_conn_cell_t p) {
  ph_conn_cell_t next;
  
  assert(conn_list != NULL);
  assert(p != NULL);

  ph_pthread_mutex_lock(&conn_list->mp);
  next = p->next;
  ph_conn_unlock(p->self);

  if (next != NULL) {
    ph_conn_lock(next->self);
  }
  ph_pthread_mutex_unlock(&conn_list->mp);

  return next;
}









