/**************************************
 * $Author: kaneda $
 * $Date: 2002/10/16 03:00:03 $
 **************************************/

/* ssh.c */

#include "ph_conn.h"


#ifndef __PH_WIN32__


ph_ssh_attr_t ph_ssh_attr_new(int send_fd, int recv_fd, int pid) {
  ph_ssh_attr_t p;
  int oldflags;

  p = (ph_ssh_attr_t)ph_malloc(sizeof(struct ph_ssh_attr));
  p->send_fd = send_fd;
  p->recv_fd = recv_fd;
  p->pid = pid;

  oldflags = ph_fcntl(p->send_fd, F_GETFD, 0);
  ph_fcntl(p->send_fd, F_SETFD, oldflags | FD_CLOEXEC);

  oldflags = ph_fcntl(p->recv_fd, F_GETFD, 0);
  ph_fcntl(p->recv_fd, F_SETFD, oldflags | FD_CLOEXEC);

  return p;
}


void ph_ssh_attr_dispose(void * p) {
  ph_ssh_attr_t attr = (ph_ssh_attr_t)p;

  if (attr == NULL) {
    ph_err_msg("ph_ssh_attr_dispose() : NULL input");
    return;
  }

  ph_free(attr);
}



static void redir_child_process(int l2r_fds[], int r2l_fds[]) {
  /* redirect stdin to l2r_fd[0] */
  close(0);
  dup(l2r_fds[0]);
  close(l2r_fds[0]);

  /* redirect stdout to r2l_fd[1] */
  close(1);
  dup(r2l_fds[1]);
  close(r2l_fds[1]); 

  /* close useless fds */
  close(l2r_fds[1]);
  close(r2l_fds[0]);
}


static void execute_ssh(char * host, char * service, char * user) {
  /* TODO: ssh $B$O0l<oN`$H$O$+$.$i$J$$(B */

  execlp("ssh",
	 "ssh", "-l", user, "-q", host, "sock", "localhost", service, (char *)0);
}


static int check_ssh(int fd, pid_t pid) {
  char s[3];   
  fd_set r_fdset;
  struct timeval 	timeout;
  ssize_t n;

  FD_ZERO(&r_fdset);
  FD_SET(fd, &r_fdset);

  timeout.tv_sec = PH_SSH_TIMEOUT;
  timeout.tv_usec = 0;
  ph_select(64, &r_fdset, NULL, NULL, &timeout);
  
  if (FD_ISSET(fd, &r_fdset) == 0) {
    return 0;
  }

  
  ph_mzero(s, 3);   
  n = read(fd, s, 3);
  return (ph_streq("OK", s));
}


void * ph_ssh_attr_connect(ph_addr_t addr) {
  int l2r_fds[2];	/* fds[1]@local  -----> fds[0]@remote */
  int r2l_fds[2];	/* fds[0]@remote <----- fds[1]@local */
  int pid;

  ph_pipe(l2r_fds);
  ph_pipe(r2l_fds);

  if ((pid = ph_fork()) == 0) {
    // child process side
    
    // to avoid ssh-askpass
    putenv("DISPLAY=");

    redir_child_process(l2r_fds, r2l_fds);
    //close(fileno(stderr));
    execute_ssh(addr->host, addr->service, addr->user);
    assert(0);
  }

  // parent process side

  /* close useless fds */
  ph_close(l2r_fds[0]);
  ph_close(r2l_fds[1]);


  if (check_ssh(r2l_fds[0], pid) == 0) {
    int status;

    ph_close(l2r_fds[1]);
    ph_close(r2l_fds[0]);
    kill(pid, SIGKILL);
    ph_waitpid(pid, &status, 0);
    return NULL;
  }

  return ph_ssh_attr_new(l2r_fds[1], r2l_fds[0], pid);
}


void * ph_ssh_attr_listen(ph_addr_t addr) {
  assert(0);
}

void * ph_ssh_attr_accept(void * p, struct sockaddr * cliaddr_p, socklen_t * len_p) {
  assert(0);
}


static ssize_t writen(int fd, void * buf, size_t len) {
  ssize_t n = 0, m;
 
  while (((m = write(fd, buf+n, len - n)) > 0) && ((n += m) < len)) {
  }

  return n;
}

ssize_t ph_ssh_attr_send(void * p, void * buf, size_t len) {
  ph_ssh_attr_t attr = (ph_ssh_attr_t)p;

  assert(attr != NULL);
  return writen(attr->send_fd, buf, len);
}

static ssize_t readn(int fd, char * buf, size_t len) {
  ssize_t n = 0, m;
 
  while (((m = read(fd, buf+n, len - n)) > 0) && ((n += m) < len)) {
  }

  return n;
}

ssize_t ph_ssh_attr_recv(void * p, void * buf, size_t len) {
  ph_ssh_attr_t attr = (ph_ssh_attr_t)p;

  assert(attr != NULL);
  return readn(attr->recv_fd, buf, len);
}


void ph_ssh_attr_close(void * p) {
  ph_ssh_attr_t attr = (ph_ssh_attr_t)p;
  int status;

  assert(attr != NULL);

  assert(attr->send_fd > 0);
  assert(attr->recv_fd > 0);
  assert(attr->pid > 0);

  ph_close(attr->send_fd);
  ph_close(attr->recv_fd);

  ph_waitpid(attr->pid, &status, 0);
}


int ph_ssh_attr_fd_isset(void * p, fd_set * fdset_p, int type) {
  ph_ssh_attr_t attr = (ph_ssh_attr_t)p;
  int n = 0;

  assert(attr != NULL); 
  assert(fdset_p != NULL);

  switch (type) {
  case PH_FD_TYPE_READ:
    n = ((attr->recv_fd >= 0) && (FD_ISSET(attr->recv_fd, fdset_p)));
    break;
  case PH_FD_TYPE_WRITE:
    n = ((attr->send_fd >= 0) && (FD_ISSET(attr->send_fd, fdset_p)));
    break;
  case PH_FD_TYPE_ERROR:
    n = (((attr->send_fd >= 0) && (FD_ISSET(attr->send_fd, fdset_p))) || 
	 ((attr->recv_fd >= 0) && (FD_ISSET(attr->recv_fd, fdset_p))));
    break;
  default:
    ph_err_quit("ph_ssh_attr_fd_isset() : no matching");
    break;
  }

  return n;
}


void ph_ssh_attr_fd_set(void * p, fd_set * fdset_p, int type) {
  ph_ssh_attr_t attr = (ph_ssh_attr_t)p;

  assert(attr != NULL); 
  assert(fdset_p != NULL);

  switch (type) {
  case PH_FD_TYPE_READ:
    FD_SET(attr->recv_fd, fdset_p);
    break;
  case PH_FD_TYPE_WRITE:
    FD_SET(attr->send_fd, fdset_p);
    break;
  case PH_FD_TYPE_ERROR:
    FD_SET(attr->send_fd, fdset_p);
    FD_SET(attr->recv_fd, fdset_p);
    break;
  default:
    ph_err_quit("ph_ssh_attr_fd_set() : no matching");
    break;
  }
}


#else


ph_ssh_attr_t ph_ssh_attr_new(int send_fd, int recv_fd, int pid) {
  ph_err_quit("ph_ssh_attr_new() : SSH is not supported in Windows");
  return NULL;
}

void ph_ssh_attr_dispose(void * p) {
  ph_err_quit("ph_ssh_attr_dispose() : SSH is not supported in Windows");
}

void * ph_ssh_attr_connect(ph_addr_t addr) {
  ph_err_quit("ph_ssh_attr_connect() : SSH is not supported in Windows");
  return NULL;
}

void * ph_ssh_attr_listen(ph_addr_t addr) {
  ph_err_quit("ph_ssh_attr_listen() : SSH is not supported in Windows");
  return NULL;
}

void * ph_ssh_attr_accept(void * p, struct sockaddr * cliaddr_p, socklen_t * len_p) {
  ph_err_quit("ph_ssh_attr_accept() : SSH is not supported in Windows");
  return NULL;
}

ssize_t ph_ssh_attr_send(void * p, void * buf, size_t len) {
  ph_err_quit("ph_ssh_attr_send() : SSH is not supported in Windows");
  return -1;
}

ssize_t ph_ssh_attr_recv(void * p, void * buf, size_t len) {
  ph_err_quit("ph_ssh_attr_recv() : SSH is not supported in Windows");
  return -1;
}

void ph_ssh_attr_close(void * p) {
  ph_err_quit("ph_ssh_attr_close() : SSH is not supported in Windows");
}

int ph_ssh_attr_fd_isset(void * p, fd_set * fdset_p, int type) {
  ph_err_quit("ph_ssh_attr_fd_isset() : SSH is not supported in Windows");
  return -1;
}

void ph_ssh_attr_fd_set(void * p, fd_set * fdset_p, int type) {
  ph_err_quit("ph_ssh_attr_fd_set() : SSH is not supported in Windows");
}


#endif
