/**************************************
 * $Author: kaneda $
 * $Date: 2003/02/12 12:45:20 $
 **************************************/

/* vp.c */
#include "ph_vp.h"


/* Public Function Prototypes */


ph_vp_range_t ph_vp_range_new(ph_vp_t lower, ph_vp_t upper) {
  ph_vp_range_t vp_range;
  
  assert(lower <= upper);
  vp_range = (ph_vp_range_t)ph_malloc(sizeof(struct ph_vp_range));
  vp_range->lower = lower;
  vp_range->upper = upper;
  
  return vp_range;
}


void ph_vp_range_dispose(ph_vp_range_t vp_range) {
  if (vp_range == NULL) {
    ph_err_msg("ph_vp_range_dispose() : NULL input");
    return;
  }
  ph_free(vp_range);
}


ph_vp_range_t ph_vp_range_dup(ph_vp_range_t vp_range) {
  return ph_vp_range_new(vp_range->lower, vp_range->upper);
}


int ph_vp_range_is_sub(ph_vp_range_t p, ph_vp_range_t q) {
  return ((q->lower <= p->lower) && (p->upper <= q->upper));
}


int ph_vp_range_is_intersected(ph_vp_range_t p, ph_vp_range_t q) {
  return ((p->upper >= q->lower) && (q->upper >= p->lower));
}


int ph_vp_range_is_exclusive(ph_vp_range_t p, ph_vp_range_t q) {
  assert(p != NULL);
  assert(q != NULL);
  return ((p->upper < q->lower) || (q->upper < p->lower));
}


int ph_vp_range_is_member(ph_vp_range_t p, ph_vp_t vp) {
  assert(p != NULL);
  return ((p->lower <= vp) && (vp <= p->upper));
}




PH_DEFINE_LIST_FUNC_BODY(ph_vp_range) 


ph_vp_range_list_t ph_vp_range_list_dup(ph_vp_range_list_t vp_range_list) {
  ph_vp_range_list_t rt;
  ph_vp_range_cell_t p;

  rt = ph_vp_range_list_new(0);
  for (p = vp_range_list->head; p != NULL; p = p->next) {
    ph_vp_range_list_enqueue(rt, ph_vp_range_dup(p->self));
  }
  return rt;
}



ph_vp_set_t ph_vp_set_new() {
  return ph_vp_range_list_new(0);
}


void ph_vp_set_dispose(ph_vp_set_t vp_set) {
  if (vp_set == NULL) {
    ph_err_msg("ph_vp_set_dispose() : NULL input");
    return;
  }

  ph_vp_range_list_dispose(vp_set);
}


static void ph_vp_set_print_sub(ph_vp_set_t vp_set, void * arg) {
  ph_vp_range_cell_t p;
  
  for (p = vp_set->head; p != NULL; p = p->next) {
    ph_debug_print("("  PH_LLD "," PH_LLD ")%s", p->self->lower, p->self->upper, ((p->next != NULL) ? "," : ""));
  }
}


void ph_vp_set_print(ph_vp_set_t vp_set) {
  assert(vp_set != NULL);
  ph_debug_print2("[vp_set ");
  ph_vp_set_print_sub(vp_set, NULL);
  ph_debug_print2("]\n");
}


ph_vp_set_t ph_vp_set_dup(ph_vp_set_t vp_set) {
   assert(vp_set != NULL);
   return ph_vp_range_list_dup(vp_set);
}

int ph_vp_set_is_empty(ph_vp_set_t vp_set) {
  assert(vp_set != NULL);
  return ph_vp_range_list_is_empty(vp_set);
}


/* p is q's subset <===> 
   \forall i \in p, \forall j \in q, if (i and j is intersected) then (i is j's subset) */
int ph_vp_set_is_sub(ph_vp_set_t p, ph_vp_set_t q) {
  int n;
  ph_vp_range_cell_t i, j;
  
  assert(p != NULL);
  assert(q != NULL);
  
  n = 1;
  i = p->head;
  j = q->head;
  while (i != NULL) {
    if (j == NULL) {
      n = 0;
      break;
    }
    
    if (ph_vp_range_is_exclusive(i->self, j->self)) {
      j = j->next;
      continue;
    }
    
    if (ph_vp_range_is_sub(i->self, j->self) == 0) {
       n = 0;
       break;
    }
    
    i = i->next;
  }
  return n;
}


int ph_vp_set_is_equal(ph_vp_set_t p, ph_vp_set_t q) {
  return (ph_vp_set_is_sub(p, q) && ph_vp_set_is_sub(q, p));
}


int ph_vp_set_is_intersected(ph_vp_set_t p, ph_vp_set_t q) {
  return !ph_vp_set_is_exclusive(p,q);
}

int ph_vp_set_is_exclusive(ph_vp_set_t p, ph_vp_set_t q) {
  int n;
  ph_vp_range_cell_t i, j;
  
  assert(p != NULL);
  assert(q != NULL);
  
  n = 1;
  i = p->head;
  j = q->head;
  while ((i != NULL) && (j != NULL)) {
    if ((i->self->upper < j->self->lower)) { 			/* i is lesser than j */
      i = i->next;
    } else if ((i->self->lower > j->self->upper)) {		/* j is lesser than i */
      j = j->next;
    } else {
      assert(ph_vp_range_is_intersected(i->self, j->self));
      n = 0;
      break;
    }
  }
  
  return n;
}


int ph_vp_set_is_member(ph_vp_set_t vp_set, ph_vp_t vp) {
  int n;
  ph_vp_range_cell_t i;

  assert(vp_set != NULL);
  n = 0;
  for (i = vp_set->head; i != NULL; i = i->next) {
    if (ph_vp_range_is_member(i->self, vp)) {
      n = 1;
      break;
    }
  }
  return n;
}


ph_vp_set_t ph_vp_set_make_range(ph_vp_t l, ph_vp_t u) {
  ph_vp_set_t vp_set;

  vp_set = ph_vp_set_new();
  ph_vp_range_list_enqueue(vp_set, ph_vp_range_new(l, u));
  return vp_set;
}


ph_vp_set_t ph_vp_set_make_empty() {
  return ph_vp_set_new();
}


ph_vp_set_t ph_vp_set_make_singleton(ph_vp_t vp) {
  ph_vp_set_t vp_set;

  vp_set = ph_vp_set_new();
  ph_vp_range_list_enqueue(vp_set, ph_vp_range_new(vp, vp));
  return vp_set;
}

ph_vp_set_t ph_vp_set_make_intersection(ph_vp_set_t p, ph_vp_set_t q) {
  ph_vp_set_t r;
  ph_vp_range_cell_t i, j;
  
  assert(p != NULL);
  assert(q != NULL);

  r = ph_vp_set_new();

  i = p->head;
  j = q->head;
  while ((i != NULL) && (j != NULL)) {
  if (i == NULL) {
      j = j->next;
      continue;
    } 

    if (j == NULL) {
      i = i->next;
      continue;
    }
    
    if (i->self->upper < j->self->lower) { 	/* i is lesser than j */
      i = i->next;
      continue;
    }
    
    if (i->self->lower > j->self->upper) { 	/* i is bigger than j */
      j = j->next;
      continue;
    }

    {							/* i and j are intersected */
      ph_vp_t l;
      ph_vp_t u;
      
      l = ph_max(i->self->lower, j->self->lower);
      u = ph_min(i->self->upper, j->self->upper);
      //printf("i=%lld,%lld\n", i->self->lower, i->self->upper);
      //printf("j=%lld,%lld\n", j->self->lower, j->self->upper);
      //printf("l,u=%lld,%lld\n", l,u);

      ph_vp_range_list_enqueue(r, ph_vp_range_new(l, u));

      if (i->self->upper == j->self->upper) {
	i = i->next;
	j = j->next;
      } else if (i->self->upper < j->self->upper) {
	i = i->next;
      } else if (j->self->upper < i->self->upper) {
	j = j->next;
      }
    }
  }

  return r;
}


ph_vp_set_t ph_vp_set_make_union(ph_vp_set_t p, ph_vp_set_t q) {
  ph_vp_set_t r;
  ph_vp_range_cell_t i, j;

  assert(p != NULL);
  assert(q != NULL);

  r = ph_vp_set_new();

  i = p->head;
  j = q->head;
  while ((i != NULL) || (j != NULL)) {
    if (i == NULL) {
      ph_vp_range_list_enqueue(r, ph_vp_range_dup(j->self));
      j = j->next;
      continue;
    } 

    if (j == NULL) {
      ph_vp_range_list_enqueue(r, ph_vp_range_dup(i->self));
      i = i->next;
      continue;
    }
    
    if ((i->self->upper < j->self->lower) && 
	(j->self->lower - i->self->upper > ph_get_ph_vp_t(1))) { 	/* i is lesser than j */
      ph_vp_range_list_enqueue(r, ph_vp_range_dup(i->self));
      i = i->next;
      continue;
    }
    
    if ((i->self->lower > j->self->upper) && 
	(i->self->lower - j->self->upper > ph_get_ph_vp_t(1))) { 	/* i is lesser than j */
      ph_vp_range_list_enqueue(r, ph_vp_range_dup(j->self));
      j = j->next;
      continue;
    }

    {							/* i and j are intersected */
      ph_vp_t l;
      ph_vp_t u;
      
      l = ph_min(i->self->lower, j->self->lower);
      u = ph_max(i->self->upper, j->self->upper);
      
      /* merge */
      i = i->next;
      j = j->next;
      while (1) {
	if ((i != NULL) && (ph_vp_range_is_member(i->self, u + ph_get_ph_vp_t(1)))) {
	  u = i->self->upper;
	  i = i->next;
	  continue;
	} 

	if ((j != NULL) && (ph_vp_range_is_member(j->self, u + ph_get_ph_vp_t(1)))) {
	  u = j->self->upper;
	  j = j->next;
	  continue;
	} 

	if ((i != NULL) && (i->self->upper <= u)) {
	  i = i->next;
	  continue;
	}

	if ((j != NULL) && (j->self->upper <= u)) {
	  j = j->next;
	  continue;
	}
	
	break;
      }
      ph_vp_range_list_enqueue(r, ph_vp_range_new(l, u));
    }
  } 

  return r;
}


ph_vp_set_t ph_vp_set_make_diff(ph_vp_set_t p, ph_vp_set_t q) {
  ph_vp_set_t r;
  ph_vp_range_cell_t i, j;
  ph_vp_t l, u;

  assert(p != NULL);
  assert(q != NULL);

  r = ph_vp_set_new();
  
  i = p->head;
  j = q->head;
  l = (i != NULL) ? i->self->lower : 0LL;
  u = (i != NULL) ? i->self->upper : 0LL;
  while (i != NULL) {

    if (j == NULL) {
      ph_vp_range_list_enqueue(r, ph_vp_range_new(l, u));
      i = i->next;
      l = (i != NULL) ? i->self->lower : 0LL;
      u = (i != NULL) ? i->self->upper : 0LL;
      continue;
    }

    if (u < j->self->lower) { 	/* [l,u] is lesser than j */
      ph_vp_range_list_enqueue(r, ph_vp_range_new(l, u));

      l = u + ph_get_ph_vp_t(1);
      u = i->self->upper;
      
      if (l > u) {
	i = i->next;
	l = (i != NULL) ? i->self->lower : 0LL;
	u = (i != NULL) ? i->self->upper : 0LL;
      }
      continue;
    } 

    if (l > j->self->upper) {		/* j is lesser than [l.u] */
      j = j->next;
      continue;
    }


    /* Then, [l,u] and j are intersected */

    if ((j->self->lower <= l) && (l <= j->self->upper)) {
      if (j->self->upper >= u) {
	i = i->next;
	l = (i != NULL) ? i->self->lower : 0LL;
	u = (i != NULL) ? i->self->upper : 0LL;
      } else {
	l = j->self->upper + ph_get_ph_vp_t(1);
	assert(l <= u);
      }

      continue;
    }
    
    if ((l <= j->self->lower) && (j->self->lower <= u)) {
      u = j->self->lower - ph_get_ph_vp_t(1);
      continue;
    }
  }

  return r;
}


ph_vp_t ph_vp_set_get_min_elem(ph_vp_set_t vp_set) {
  assert(vp_set != NULL);
  return (ph_vp_range_list_is_empty(vp_set)) ? PH_INVALID_VP : vp_set->head->self->lower;
}


ph_vp_t ph_vp_set_get_max_elem(ph_vp_set_t vp_set) {
  assert(vp_set != NULL);
  return (ph_vp_range_list_is_empty(vp_set)) ? PH_INVALID_VP : vp_set->tail->self->upper;
}

static ph_vp_t ph_vp_set_get_nelm(ph_vp_set_t vp_set) {
  ph_vp_t n;
  ph_vp_range_cell_t p;

  n = ph_get_ph_vp_t(0);
  for (p = vp_set->head; p != NULL; p = p->next) {
    n += p->self->upper - p->self->lower + ph_get_ph_vp_t(1);
  }
  return n;
}

ph_vp_t ph_vp_set_get_any_elem(ph_vp_set_t vp_set) {
  ph_vp_t vp, n, r;
  ph_vp_range_cell_t p;

  assert(vp_set != NULL);
  if (ph_vp_range_list_is_empty(vp_set)) {
    return PH_INVALID_VP;
  }

  n = ph_vp_set_get_nelm(vp_set);
  
  r = ((ph_vp_t)(lrand48()) << 31) + (ph_vp_t)(lrand48());
  if (r >= PHOENIX_VP_LIMIT) {
    r = PHOENIX_MAX_VP;
  }
  assert((PH_MIN_VP <= vp) && (vp <= PHOENIX_MAX_VP));

  r = r % n;

  vp = ph_get_ph_vp_t(0);
  for (p = vp_set->head; p != NULL; p = p->next) {
    if (r <  p->self->upper - p->self->lower + ph_get_ph_vp_t(1)) {
      vp = p->self->lower + r;
      break;
    }
    r -= p->self->upper - p->self->lower + ph_get_ph_vp_t(1);
  }

  return vp;
}


void ph_vp_set_add_range(ph_vp_set_t * vp_set_p, ph_vp_t lower, ph_vp_t upper) {
  ph_vp_set_t p, q;

  assert(vp_set_p != NULL);
  assert((*vp_set_p) != NULL);

  p = ph_vp_set_make_range(lower, upper);
  q = ph_vp_set_make_union((*vp_set_p), p);
  ph_vp_set_dispose((*vp_set_p));
  ph_vp_set_dispose(p);
  (*vp_set_p) = q;
}

void ph_vp_set_remove_range(ph_vp_set_t * vp_set_p, ph_vp_t lower, ph_vp_t upper) {
  ph_vp_set_t p, q;

  assert(vp_set_p != NULL);
  assert((*vp_set_p) != NULL);

  p = ph_vp_set_make_range(lower, upper);
  q = ph_vp_set_make_diff((*vp_set_p), p);
  ph_vp_set_dispose((*vp_set_p));
  ph_vp_set_dispose(p);
  (*vp_set_p) = q;
}


void ph_vp_set_vps2str(ph_vp_set_t vp_set, char * s) {
  ph_vp_range_cell_t p;
  int i;
  int num;

  assert(vp_set != NULL);  
  assert(s != NULL);

  num = 0;
  for (p = vp_set->head, i = 0; p != NULL; p = p->next, i++) {
    snprintf(s + (i+1)*50, 50, PH_LLD " " PH_LLD, p->self->lower, p->self->upper);
    num++;
  }
  snprintf(s, 50, "%d", num);
}

ph_vp_set_t ph_vp_set_str2vps(char * s) {
  ph_vp_set_t vp_set;
  int num, i;

  assert(s != NULL);

  vp_set = ph_vp_set_make_empty();
  sscanf(s, "%d", &num);
  for (i = 0; i < num; i++) {
    ph_vp_t l, u;
    sscanf(s + (i+1)*50, PH_LLD " " PH_LLD, &l, &u);
    ph_vp_set_add_range(&vp_set, l, u);
  }
  return vp_set;
}




void ph_vp_set_vps2str_for_print(ph_vp_set_t vp_set, char * s) {
  ph_vp_range_cell_t p;
  assert(vp_set != NULL);  
  assert(s != NULL);

  for (p = vp_set->head; p != NULL; p = p->next) {
    snprintf(s + ph_strlen(s), 50, 
	     "[" PH_LLD "," PH_LLD ")", 
	     p->self->lower, p->self->upper + ph_get_ph_vp_t(1));
  }
}


ph_vp_set_sync_t ph_vp_set_sync_new() {
  ph_vp_set_sync_t p;

  p = (ph_vp_set_sync_t)ph_malloc(sizeof(struct ph_vp_set_sync));
  p->vp_set = ph_vp_set_new();
  ph_pthread_mutex_init(&p->mp, NULL);  
  return p;
}


void ph_vp_set_sync_dispose(ph_vp_set_sync_t vp_set_sync) {
  if (vp_set_sync == NULL) {
    ph_err_msg("ph_vp_set_sync_dispose() : NULL input");
    return;
  }
  ph_vp_set_dispose(vp_set_sync->vp_set);
  ph_pthread_mutex_destroy(&vp_set_sync->mp);
  ph_free(vp_set_sync);
}


void ph_vp_set_sync_assume(ph_vp_set_sync_t p, ph_vp_set_t q) {
  ph_vp_set_t r;

  assert(p != NULL);
  assert(q != NULL);
  
  ph_pthread_mutex_lock(&p->mp);
  r = ph_vp_set_make_union(p->vp_set, q);
  ph_vp_range_list_swap(p->vp_set, r, 0);
  ph_vp_set_dispose(r);

  ph_pthread_mutex_unlock(&p->mp);
}

void ph_vp_set_sync_release(ph_vp_set_sync_t p, ph_vp_set_t q) {
  ph_vp_set_t r;

  assert(p != NULL);
  assert(q != NULL);
  
  ph_pthread_mutex_lock(&p->mp);
  r = ph_vp_set_make_diff(p->vp_set, q);
  ph_vp_range_list_swap(p->vp_set, r, 0);
  ph_vp_set_dispose(r);

  ph_pthread_mutex_unlock(&p->mp);
}


ph_vp_set_t ph_vp_set_sync_dup(ph_vp_set_sync_t vp_set_sync) {
  ph_vp_set_t vp_set;

  assert(vp_set_sync != NULL);
  ph_pthread_mutex_lock(&vp_set_sync->mp);
  vp_set = ph_vp_set_dup(vp_set_sync->vp_set);
  ph_pthread_mutex_unlock(&vp_set_sync->mp);
  return vp_set;
}


int ph_vp_set_sync_is_member(ph_vp_set_sync_t vp_set_sync, ph_vp_t vp) {
  int n;

  assert(vp_set_sync != NULL);
  ph_pthread_mutex_lock(&vp_set_sync->mp);
  n = ph_vp_set_is_member(vp_set_sync->vp_set, vp);
  ph_pthread_mutex_unlock(&vp_set_sync->mp);
  return n;
}


ph_vp_t ph_vp_set_sync_get_max_elem(ph_vp_set_sync_t vp_set_sync) {
  ph_vp_t vp;

  assert(vp_set_sync != NULL);
  ph_pthread_mutex_lock(&vp_set_sync->mp);
  vp = ph_vp_set_get_max_elem(vp_set_sync->vp_set);
  ph_pthread_mutex_unlock(&vp_set_sync->mp);
  return vp;
}



