/*************************************
 * $Author: kaneda $
 * $Date: 2002/11/13 07:29:10 $
 **************************************/

/* wrap_unix.c */

#include "ph_util.h"



void ph_free(void *ptr) {
  free(ptr);
}

void * ph_malloc(size_t size) {
  void *ptr;

  if ((ptr = malloc(size)) == NULL) {
    ph_err_sys("malloc error");
  }
	
  return(ptr);
}


void * ph_calloc(size_t n, size_t size) {
  void *ptr;

  if ((ptr = calloc(n, size)) == NULL) {
    ph_err_sys("calloc error");
  }

  return(ptr);
}



#ifndef __PH_WIN32__
void ph_close(int fd) {
  if (close(fd) == -1) {
    ph_err_msg("close error");
  }
}


void ph_dup(int fd) {
  if (dup(fd) == -1) {
    ph_err_sys("dup error");
  }
}


void ph_dup2(int fd1, int fd2) {
  if (dup2(fd1, fd2) == -1) {
    ph_err_sys("dup2 error");
  }
}

int ph_fcntl(int fd, int cmd, int arg) {
  int n;
  
  if ((n = fcntl(fd, cmd, arg)) == -1) {
    ph_err_sys("fcntl error");
  }

  return(n);
}

void ph_gettimeofday(struct timeval *tv, void *foo) {
  if (gettimeofday(tv, foo) == -1) {
    ph_err_sys("gettimeofday error");
  }

  return;
}

char * ph_getenv(const char *name) {
  char * s;

  if ((s = getenv(name)) != NULL) {
    return s;
  }

  ph_err_msg("getenv error");
  return NULL;
}

int ph_ioctl(int fd, int request, void *arg) {
  int n;

  if ((n = ioctl(fd, request, arg)) == -1) {
    ph_err_sys("ioctl error");
  }

  return(n);	/* streamio of I_LIST returns value */
}


pid_t ph_fork(void) {
  pid_t	pid;

  if ((pid = fork()) == -1) {
    ph_err_sys("fork error");
  }

  return(pid);
}


void ph_mkstemp(char *temp) {
  if (mkstemp(temp) == -1 || temp[0] == 0) {
    ph_err_quit("mkstemp error");
  }
}

int ph_open(const char *pathname, int oflag, mode_t mode) {
  int fd;

  if ((fd = open(pathname, oflag, mode)) == -1) {
    ph_err_sys("open error for %s", pathname);
  }

  return(fd);
}

void ph_pipe(int *fds) {
  if (pipe(fds) < 0) {
    ph_err_sys("pipe error");
  }
}

ssize_t ph_read(int fd, void *ptr, size_t nbytes) {
  ssize_t n;

  if ((n = read(fd, ptr, nbytes)) == -1) {
    ph_err_sys("read error");
  }
	
  return(n);
}

void ph_sigaddset(sigset_t *set, int signo) {
  if (sigaddset(set, signo) == -1) {
    ph_err_sys("sigaddset error");
  }
}

void ph_sigdelset(sigset_t *set, int signo) {
  if (sigdelset(set, signo) == -1) {
    ph_err_sys("sigdelset error");
  }
}

void ph_sigemptyset(sigset_t *set) {
  if (sigemptyset(set) == -1) {
    ph_err_sys("sigemptyset error");
  }
}

void ph_sigfillset(sigset_t *set) {
  if (sigfillset(set) == -1) {
    ph_err_sys("sigfillset error");
  }
}

int ph_sigismember(const sigset_t *set, int signo) {
  int n;
  
  if ((n = sigismember(set, signo)) == -1) {
    ph_err_sys("sigismember error");
  }
  
  return(n);
}

void ph_sigpending(sigset_t *set) {
  if (sigpending(set) == -1) {
    ph_err_sys("sigpending error");
  }
}

void ph_sigprocmask(int how, const sigset_t *set, sigset_t *oset) {
  if (sigprocmask(how, set, oset) == -1) {
    ph_err_sys("sigprocmask error");
  }
}

long ph_sysconf(int name) {
  long val;

  errno = 0;		/* in case sysconf() does not change this */
  if ((val = sysconf(name)) == -1) {
    ph_err_sys("sysconf error");
  }

  return(val);
}

void ph_unlink(const char *pathname) {
  if (unlink(pathname) == -1) {
    ph_err_sys("unlink error for %s", pathname);
  }
}

pid_t ph_wait(int *iptr) {
  pid_t	pid;

  if ((pid = wait(iptr)) == -1) {
    ph_err_sys("wait error");
  }
  
  return pid;
}

pid_t ph_waitpid(pid_t pid, int *iptr, int options) {
  pid_t	retpid;

  if ((retpid = waitpid(pid, iptr, options)) == -1) {
    ph_err_sys("waitpid error");
  }
  
  return retpid;
}

void ph_write(int fd, void *ptr, size_t nbytes) {
  if (write(fd, ptr, nbytes) != nbytes) {
    ph_err_sys("write error");
  }
}

FILE * ph_fopen(const char * filename, const char * mode) {
  FILE * fp;

  assert(filename != NULL);
  assert(mode != NULL);

  if ((fp = fopen(filename, mode)) == NULL) {
    ph_err_sys("fopen error");    
  }

  return fp;
}

void ph_fclose(FILE * fp) {
  assert(fp != NULL);

  if (fclose(fp) != 0) {
    ph_err_sys("fclose error");
  }
}
			 

ph_sigfunc * ph_signal(int signo, ph_sigfunc * func) {
  struct sigaction      act, oact;
  
  act.sa_handler = func;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  if (signo == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT;       /* SunOS 4.x    */
#endif
  } else {
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART;         /* SVR4, 44BSD */
#endif
  }
  
  if (sigaction(signo, &act, &oact) < 0) {
    return SIG_ERR;
  }
  return oact.sa_handler;
}

#else

#endif



