/*************************************
 * $Author: kaneda $
 * $Date: 2002/12/20 10:14:41 $
 **************************************/

/* wrap_unix.c */

#include "util.h"


void Free(void *ptr) {
  free(ptr);
}

void * Malloc(size_t size) {
  void *ptr;

  if ((ptr = malloc(size)) == NULL) {
    err_sys("malloc error");
  }
	
  return(ptr);
}


void * Calloc(size_t n, size_t size) {
  void *ptr;

  if ((ptr = calloc(n, size)) == NULL) {
    err_sys("calloc error");
  }

  return(ptr);
}




void Close(int fd) {
  if (close(fd) == -1) {
    err_msg("close error");
  }
}


void Dup(int fd) {
  if (dup(fd) == -1) {
    err_sys("dup error");
  }
}


void Dup2(int fd1, int fd2) {
  if (dup2(fd1, fd2) == -1) {
    err_sys("dup2 error");
  }
}

int Fcntl(int fd, int cmd, int arg) {
  int n;
  
  if ((n = fcntl(fd, cmd, arg)) == -1) {
    err_sys("fcntl error");
  }

  return(n);
}

void Gettimeofday(struct timeval *tv, void *foo) {
  if (gettimeofday(tv, foo) == -1) {
    err_sys("gettimeofday error");
  }

  return;
}

char * Getenv(const char *name) {
  char * s;

  if ((s = getenv(name)) != NULL) {
    return s;
  }

  err_msg("getenv error");
  return NULL;
}

int Ioctl(int fd, int request, void *arg) {
  int n;

  if ((n = ioctl(fd, request, arg)) == -1) {
    err_sys("ioctl error");
  }

  return(n);	/* streamio of I_LIST returns value */
}


pid_t Fork(void) {
  pid_t	pid;

  if ((pid = fork()) == -1) {
    err_sys("fork error");
  }

  return(pid);
}


void Mkstemp(char *temp) {
  if (mkstemp(temp) == -1 || temp[0] == 0) {
    err_quit("mkstemp error");
  }
}

int Open(const char *pathname, int oflag, mode_t mode) {
  int fd;

  if ((fd = open(pathname, oflag, mode)) == -1) {
    err_sys("open error for %s", pathname);
  }

  return(fd);
}

void Pipe(int *fds) {
  if (pipe(fds) < 0) {
    err_sys("pipe error");
  }
}

ssize_t Read(int fd, void *ptr, size_t nbytes) {
  ssize_t n;

  if ((n = read(fd, ptr, nbytes)) == -1) {
    err_sys("read error");
  }
	
  return(n);
}

void Sigaddset(sigset_t *set, int signo) {
  if (sigaddset(set, signo) == -1) {
    err_sys("sigaddset error");
  }
}

void Sigdelset(sigset_t *set, int signo) {
  if (sigdelset(set, signo) == -1) {
    err_sys("sigdelset error");
  }
}

void Sigemptyset(sigset_t *set) {
  if (sigemptyset(set) == -1) {
    err_sys("sigemptyset error");
  }
}

void Sigfillset(sigset_t *set) {
  if (sigfillset(set) == -1) {
    err_sys("sigfillset error");
  }
}

int Sigismember(const sigset_t *set, int signo) {
  int n;
  
  if ((n = sigismember(set, signo)) == -1) {
    err_sys("sigismember error");
  }
  
  return(n);
}

void Sigpending(sigset_t *set) {
  if (sigpending(set) == -1) {
    err_sys("sigpending error");
  }
}

void Sigprocmask(int how, const sigset_t *set, sigset_t *oset) {
  if (sigprocmask(how, set, oset) == -1) {
    err_sys("sigprocmask error");
  }
}

long Sysconf(int name) {
  long val;

  errno = 0;		/* in case sysconf() does not change this */
  if ((val = sysconf(name)) == -1) {
    err_sys("sysconf error");
  }

  return(val);
}

void Unlink(const char *pathname) {
  if (unlink(pathname) == -1) {
    err_sys("unlink error for %s", pathname);
  }
}

pid_t Wait(int *iptr) {
  pid_t	pid;

  if ((pid = wait(iptr)) == -1) {
    err_sys("wait error");
  }
  
  return pid;
}

pid_t Waitpid(pid_t pid, int *iptr, int options) {
  pid_t	retpid;

  if ((retpid = waitpid(pid, iptr, options)) == -1) {
    err_sys("waitpid error");
  }
  
  return retpid;
}

void Write(int fd, void *ptr, size_t nbytes) {
  if (write(fd, ptr, nbytes) != nbytes) {
    err_sys("write error");
  }
}

FILE * Fopen(const char * filename, const char * mode) {
  FILE * fp;

  assert(filename != NULL);
  assert(mode != NULL);

  if ((fp = fopen(filename, mode)) == NULL) {
    err_sys("fopen error");    
  }

  return fp;
}

void Fclose(FILE * fp) {
  assert(fp != NULL);

  if (fclose(fp) != 0) {
    err_sys("fclose error");
  }
}
			 

Sigfunc * Signal(int signo, Sigfunc * func) {
  struct sigaction      act, oact;
  
  act.sa_handler = func;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  if (signo == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT;       /* SunOS 4.x    */
#endif
  } else {
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART;         /* SVR4, 44BSD */
#endif
  }
  
  if (sigaction(signo, &act, &oact) < 0) {
    return SIG_ERR;
  }
  return oact.sa_handler;
}





