/************************************
 * $Author: kaneda $
 * $Date: 2002/12/20 10:14:43 $
 **************************************/

/* vsh.c */

#include "vsh.h"

static int bin_vpg_show(char ** argv, char * ops, int func);
static int bin_vpg_kill(char ** argv, char * ops, int func);
static int bin_vpg_jobs(char ** argv, char * ops, int func);

enum {
  V_BUILTIN_N	= 3
};


static struct v_builtin v_builtins[] = {
  {"show", &bin_vpg_show},
  {"kill", &bin_vpg_kill},
  {"jobs", &bin_vpg_jobs}
};


vsh_t	vsh = NULL;
task_t	current_task = NULL;




vsh_t vsh_new() {
  vsh_t vsh;

  vsh = (vsh_t)Malloc(sizeof(struct vsh));
  vsh->tasks = task_list_new(0);
  vsh->machines = ph_vp_range_list_new(0);

  vsh->is_initialized = 0;

  vsh->host_vp = -1LL;

  return vsh;
}


void vsh_dispose() {
  if (vsh == NULL) {
    err_msg("vsh_dispose() : NULL input");
    return;
  }

  task_list_dispose(vsh->tasks);
  ph_vp_range_list_dispose(vsh->machines);
  Free(vsh);
}


void vsh_print() {
  debug_print2("[vsh \n");
  debug_print("fd=%d, my_vp=%lld, host_vp=%lld, id=%lld\n", vsh->fd, vsh->my_vp, vsh->host_vp, vsh->id);
  
  debug_print2("]\n");
}
	       


/**************************************************************************************
 **************************************************************************************/


task_t vsh_create_new_task() {
  task_t task;
  
  task = task_new();
  if ((vsh != NULL) && (vsh->is_initialized)) {
    task->vp = task->in_vp = task->out_vp = vsh->my_vp;
  }
  return task;
}


static int argv2str(char ** argv, char * s, size_t len) {
  int i, n;

  assert(argv != NULL);
  Mzero(s, len);
  n = 0;
  for (i = 0; argv[i] != NULL; i++) {   
    int m = Strlen(argv[i])+1;

    Mmove(s + n, argv[i], m);
    n += m;
    s[n-1] = '\0';
    assert(n <= len);
  }

  return n;
}


static void vsh_send(int fd, ph_vp_t dest_vp, task_id_t dest_id, msg_type_t msg_type, ...) {
  msg_t msg = NULL;
  va_list ap;
  
  debug_print("vsh_send()\t\t\t: fd=%d, dest_vp=%llu, dest_id=%d, msg_type=%d\n", fd, dest_vp, dest_id, msg_type);

  va_start(ap, msg_type);
  switch (msg_type) {
  case SPAWN_TASK: {
    int MSG_SIZE = 8192;
    char * body;
    task_id_t id, out_id;
    task_type_t type;
    ph_vp_t out_vp;
    char ** argv;
    int argn, n, m;
    

    id = (task_id_t)va_arg(ap, task_id_t);
    type = (task_type_t)va_arg(ap, task_type_t);
    out_vp = (ph_vp_t)va_arg(ap, ph_vp_t);
    out_id = (task_id_t)va_arg(ap, task_id_t);
    argv = (char **)va_arg(ap, char **);
    argn = (int)va_arg(ap, int);


    body = (char *)Calloc(MSG_SIZE, sizeof(char));
    Mzero(body, MSG_SIZE);
    Snprintf(body, MSG_SIZE, "%lld %d %lld %lld %d", 
	     id, (int)type, out_vp, out_id, argn);
    n = Strlen(body)+1;
    m = argv2str(argv, body + n, MSG_SIZE - n);

    msg = msg_new(dest_vp, dest_id, msg_type, body, n + m);
    break;
  }
  case JOIN: 
    msg = msg_new(dest_vp, dest_id, msg_type, Mdup("JOIN", 5), 5);
    break;
  case SET_ROUTE: {
    int MSG_SIZE = 300;
    char * body;
    task_type_t type;
    ph_vp_t vp;
    task_id_t id;
    
    type = (task_type_t)va_arg(ap, task_type_t);
    vp = (ph_vp_t)va_arg(ap, ph_vp_t);
    id = (task_id_t)va_arg(ap, task_id_t);

    body = (char *)Calloc(MSG_SIZE, sizeof(char));
    Mzero(body, MSG_SIZE);
    Snprintf(body, MSG_SIZE, "%d %lld %lld", type, vp, id);

    msg = msg_new(dest_vp, dest_id, msg_type, body, Strlen(body)+1);
    break;
  }
  case GET_MACHINE_LIST:
    msg = msg_new(dest_vp, dest_id, msg_type, Mdup("GET_MACHINE_LIST", 17), 17);
    break;
  default:
    err_quit("vsh_send() : no matching\n");
    break;
  }
  va_end(ap);

  assert(msg != NULL);
  msg_send_raw(msg, fd);
  msg_dispose(msg);
}

static void diff_machines(ph_vp_range_list_t old, ph_vp_range_list_t new) {
  ph_vp_range_cell_t p, q;

  assert(old != NULL);
  assert(new != NULL);

  p = old->head;
  q = new->head;

  /* Note : p and q are sorted. (because they are obtained by ph_get_routes().) */

  while ((p != NULL) || (q != NULL)) {

    if ((p != NULL) && (q != NULL) && (p->self->lower == q->self->lower)) {
      p = p->next;
      q = q->next;
      continue;
    }

    if ((p != NULL) && ((q == NULL) || (p->self->lower < q->self->lower))) {
      char s[BUFSIZE];
      
      vp2nickname(p->self->lower, s, BUFSIZE);
      /* Do not print message when the shell is non-interactive mode */
      if (vsh->is_interactive) { fprintf(stderr, "vpg: %s left.\n\n", s); }
      p = p->next;
      continue;
    }

    if ((q != NULL) && ((p == NULL) || (q->self->lower < p->self->lower))) {
      char s[BUFSIZE];
      
      vp2nickname(q->self->lower, s, BUFSIZE);
      if (vsh->is_interactive) { fprintf(stderr, "vpg: %s joined.\n\n", s); }
      q = q->next;
      continue;
    }
  }
}

static void vsh_recv(int fd) {
  msg_t msg;
  
  debug_print("vsh_recv()\t\t\t: fd=%d\n", fd);

  if ((msg = msg_recv_raw(fd)) == NULL) {
    return;
  }

  switch (msg->type) {
  case ASSUME_VP:
    sscanf(msg->body, "%lld %lld", &vsh->host_vp, &vsh->my_vp);
    vsh->is_initialized = 1;
    break;
  case SEND_MACHINE_LIST: {
    ph_vp_range_list_t new_machines;
    int i, nelm;

    assert(msg->body != NULL);

    sscanf(msg->body, "%d", &nelm);

    new_machines = ph_vp_range_list_new(0);
    for (i = 0; i < nelm; i++) {
      ph_vp_t lower, upper;
      sscanf(msg->body + 100*(i+1), "%lld %lld", &lower, &upper);
      ph_vp_range_list_insert(new_machines, ph_vp_range_new(lower, upper));
    }
    
    diff_machines(vsh->machines, new_machines);

    ph_vp_range_list_dispose(vsh->machines);
    vsh->machines = new_machines;
    break;
    }
  default:
    err_quit("() : no matching");
    break;
  }
  msg_dispose(msg);
}

/**************************************************************************************
 **************************************************************************************/


static int connect_to_daemon(int is_for_task, int type, ph_vp_t vp, task_id_t id) {
  int fd;
  char service[BUFSIZE];

  /* TODO */
  Mzero(service, BUFSIZE);
  Snprintf(service, BUFSIZE, "%d", DEFAULT_SH_PORT);   

  if ((fd = tcp_connect("localhost", service)) <= 0) {
    err_quit("cannt connect to local daemon (port=%s)", service);
  }

  debug_print("connect_to_daemon()\t\t: fd=%d, type=%d, vp=%lld, id=%lld\n", fd, type, vp, id);

  if (is_for_task) {
    int oldflags;

    oldflags = Fcntl(fd, F_GETFD, 0);
    Fcntl(fd, F_SETFD, oldflags | FD_CLOEXEC);
    vsh_send(fd, vsh->host_vp, 0, SET_ROUTE, type, vp, id);
  }

  return fd;
}



/**************************************************************************************
 **************************************************************************************/


static task_id_t get_fresh_id() {
  /*
  task_id_t id;

  // currently fresh id is obtained by random 
  id = (task_id_t)ph_get_random_vp();
  if (id == 0LL) { id++; }

  return id;
  */

  vsh->id++;
  return vsh->id;

}


static void release_fresh_id(task_id_t id) {
  
}


static int is_remote(char *pth) {
  return ((pth != NULL) && (strchr(pth, '@') != NULL));
}


static int is_remote_vp(ph_vp_t vp) {
  return (vp != vsh->my_vp);
}


static void partition_command(char * pth, char ** host_p) {
  strtok_r(pth, "@", host_p);
}

static char ** argv_new(LinkList args, int * argn_p) {
  int i;
  LinkNode node;
  char ** argv;

  if ((args == NULL) || (args->first == NULL)) {
    return NULL;
  }

  (*argn_p) = 0;
  for (node = firstnode(args); node; incnode(node)) {
    (*argn_p)++;
  }
  
  argv = (char **)Calloc((*argn_p)+1, sizeof(char*));
  
  for (i = 0, node = firstnode(args); node; incnode(node), i++) {
    argv[i] = Strdup(getdata(node));
  } 

  argv[(*argn_p)] = NULL;
  return argv;
}



/**************************************************************************************
 **************************************************************************************/


static char * get_host_sub(Estate state, char * s) {
  Wordcode saved_pc = state->pc;
  wordcode code;
  LinkList args;
  LinkList redir;
  int type, htok = 0;
  char * t = NULL;


  Mzero(s, BUFSIZE);

  redir = (wc_code(*state->pc) == WC_REDIR ? ecgetredirs(state) : NULL);

  code = *(state->pc ++);
  type = wc_code(code);
  
  if (type == WC_SIMPLE) { 
    if ((args = ecgetlist(state, WC_SIMPLE_ARGC(code), EC_DUP, &htok)) != NULL) {
      char * u;

      if (args && htok) {
	prefork(args, 1);
      }
    
      u = (char *)peekfirst(args);
      if (is_remote(u)) {
	Mmove(s, u, Strlen(u)+1);
	partition_command(s, &t);
      }
    }
  }
  state->pc = saved_pc;

  return t;
}

static char * get_host(Estate state, wordcode pcode, char * s) {
  char * t;

  if (WC_PIPE_TYPE(pcode) == WC_PIPE_END) { 
    t = get_host_sub(state, s);
  } else {
    Wordcode pc;
    
    state->pc++;
    for (pc = state->pc; wc_code(*pc) == WC_REDIR; pc += 3);

    t =  get_host_sub(state, s);
  }
  return t;
}

task_t vsh_setup_pipeline(Estate state, Wordcode next) {
  task_t next_task;

  if ((vsh == NULL) || (vsh->is_initialized == 0)) {
    return NULL;
  }

  assert(current_task != NULL);

  next_task = vsh_create_new_task();

  { 
    Wordcode saved_pc = state->pc;
    char s[BUFSIZE];
    char * t;

    if ((t = get_host_sub(state, s)) != NULL) {
      current_task->vp = nickname2vp(t);
    }

    state->pc = next;

    // $B<!$,(B redirection $B$N>l9g$b$"$k(B 

    if ((t = get_host(state, *state->pc++, s)) != NULL) {
      next_task->vp = nickname2vp(t);
    }

    state->pc = saved_pc;
  }
  
  if (current_task->id == 0LL) {
    current_task->id = get_fresh_id();
  }
  next_task->id = get_fresh_id();


  current_task->out_vp = next_task->vp;
  current_task->out_id = next_task->id;
  next_task->in_vp = current_task->vp;
  next_task->in_id = current_task->id;

  return next_task;
}


/**************************************************************************************
 **************************************************************************************/


void vsh_setup_task(LinkList args, LinkList redir) {
  char * host;

  if ((vsh == NULL) || (vsh->is_initialized == 0)) {
    return;
  }

  assert(current_task != NULL);

  current_task->argv = argv_new(args, &current_task->argn);
  if ((current_task->argv != NULL) && (is_remote(current_task->argv[0]))) {
    partition_command(current_task->argv[0], &host);
    current_task->vp = nickname2vp(host);
  } else {
    current_task->vp = vsh->my_vp;
  }

  if (current_task->id == 0) { 
    current_task->id = get_fresh_id();
  }


  if (redir && nonempty(redir)) {
    LinkNode node;

    for (node = firstnode(redir); node; incnode(node)) {
      Redir fn;
      char ** argv;
      ph_vp_t vp;
      task_id_t id;

      fn = (Redir)getdata(node);
      if (is_remote(fn->name) == 0) {
	continue;
      }

      argv = (char **)Calloc(2, sizeof(char*));
      argv[0] = Strdup(unmeta(fn->name));
      argv[1] = NULL;
      partition_command(argv[0], &host);
      vp = nickname2vp(host);
      id = get_fresh_id();      
      
      switch (fn->type) {
      case REDIR_READ:
	if (is_remote_vp(current_task->in_vp)) {
	  argv_dispose(argv);
	  continue;
	}
	current_task->in_argv = argv;
	current_task->in_vp = vp;
	current_task->in_id = id;
	break;
      case REDIR_WRITE:
	if (is_remote_vp(current_task->out_vp)) {
	  argv_dispose(argv);
	  continue;
	}
	current_task->out_argv = argv;
	current_task->out_vp = vp;
	current_task->out_id = id;
	break;
      default:
	err_quit("vsh_setup_task() : no matching\n");
	break;
      }
    }
  }


  if ((is_remote_vp(current_task->out_vp) == 0) && (current_task->out_id == 0)){
    current_task->out_id = get_fresh_id();
  }
  
}


/**************************************************************************************
 **************************************************************************************/

int vsh_try_remote_redirection(Redir fn) {
  int fd = -1;

  if ((vsh == NULL) || (vsh->is_initialized == 0) || (is_remote(fn->name) == 0)) {
    return -1;
  }

  assert(current_task != NULL);

  switch (fn->type) {
  case REDIR_READ:
    assert(is_remote_vp(current_task->in_vp));
    
    if (current_task->in_argv == NULL) {
      return -1;
    }

    if (is_remote_vp(current_task->vp)) {
      fd = fn->fd1;
    } else {
      fd = connect_to_daemon(1, TASK_IN, current_task->vp, current_task->id);
    }

    vsh_send(vsh->fd, current_task->in_vp, 0, SPAWN_TASK,
	     current_task->in_id, REDIR_R, current_task->vp, current_task->id, current_task->in_argv, 1);
    break;
  case REDIR_WRITE:
    assert(is_remote_vp(current_task->out_vp));

    if (current_task->out_argv == NULL) {
      return -1;
    }

    if (is_remote_vp(current_task->vp)) {
      fd = fn->fd1;
    } else {
      fd = connect_to_daemon(1, TASK_OUT, current_task->out_vp, current_task->out_id);
    }

    vsh_send(vsh->fd, current_task->out_vp, 0, SPAWN_TASK, 
	     current_task->out_id, REDIR_W, vsh->my_vp, 0LL, current_task->out_argv, 1);
    break;
  default:
    err_quit("vsh_try_remote_redirection() : no matching\n");
    break;
  }

  return fd;
}


/**************************************************************************************
 **************************************************************************************/


static void spawn_remote_task(int fds[], int is_finished[]) {
  assert(current_task != NULL);

  is_finished[0] = is_finished[1] = 0;

  if (is_remote_vp(current_task->in_vp)) {    	/* both command and input redirection are remote  */
    is_finished[0] = 1;
  } else {				  	/* command is remote and input redirection is local  */
    fds[0] = connect_to_daemon(1, TASK_OUT, current_task->vp, current_task->id);
  }

  if (is_remote_vp(current_task->out_vp)) {	/* both command and output redirection are remote  */
    is_finished[1] = 1;
  } else {					/* command is remote and output redirection is local  */
    fds[1] = connect_to_daemon(1, TASK_IN, current_task->out_vp, current_task->out_id);
  }


  vsh_send(vsh->fd, current_task->vp, 0, SPAWN_TASK, 
	   current_task->id, EXEC, current_task->out_vp, current_task->out_id, 
	   current_task->argv, current_task->argn);
}


static void initialize_select(fd_set * r_fdset_p, fd_set * w_fdset_p, fd_set * e_fdset_p, int fds[], int is_finished[]) {
  FD_ZERO(r_fdset_p);
  FD_ZERO(w_fdset_p);
  FD_ZERO(e_fdset_p);

  if (is_finished[0] == 0) {
    FD_SET(0, r_fdset_p);
    FD_SET(0, e_fdset_p);
  }

  if (is_finished[1] == 0) {
    FD_SET(fds[1], r_fdset_p);
    FD_SET(fds[1], e_fdset_p);
  }
}


static void handle_io(fd_set * r_fdset_p, fd_set * w_fdset_p, fd_set * e_fdset_p, int fds[], int is_finished[]) {
  char s[BUFSIZE];
  ssize_t n;

  if (is_finished[0] == 0) {
    if (FD_ISSET(0, r_fdset_p)) {
      ph_mzero(s, BUFSIZE);
      if ((n = read(0, s, BUFSIZE)) <= 0) {
	is_finished[0] = 1;
	Shutdown(fds[0], 0);
	Close(fds[0]);
      } else {
	send(fds[0], s, n, 0);
      }
    }
    
    if (FD_ISSET(0, e_fdset_p)) {
      is_finished[0] = 1;
      Shutdown(fds[0], 0);
      Close(fds[0]);
    }
  }


  if (is_finished[1] == 0) {
    if (FD_ISSET(fds[1], r_fdset_p)) {
      if ((n = recv(fds[1], s, PH_BUFSIZE, 0)) <= 0) {
	is_finished[1] = 1;
	Shutdown(fds[1], 0);
	Close(fds[1]);
	Close(1);
      } else {
	write(1, s, n);
      }
    }

    if (FD_ISSET(fds[1], e_fdset_p)) {
      is_finished[1] = 1;
      Shutdown(fds[1], 0);
      Close(fds[1]);
      Close(1);
    }
  }
}


int vsh_try_spawn_remote_task(char * pth, char ** argv){ 
  int fds[3], is_finished[3];
  if ((vsh == NULL) || (vsh->is_initialized == 0)) {
    return 0;
  }

  assert(current_task != NULL);

  if (is_remote_vp(current_task->vp) == 0) {
    return 0;
  }

  spawn_remote_task(fds, is_finished);

  // $B%8%g%V$,=*N;$9$k$^$G7+$jJV$9(B          TODO: 
  //while ((is_finished[0] == 0) || (is_finished[1] == 0)) {
  while (is_finished[1] == 0) {
    fd_set r_fdset, w_fdset, e_fdset;

    initialize_select(&r_fdset, &w_fdset, &e_fdset, fds, is_finished);
    select(MAX_FD, &r_fdset, &w_fdset, &e_fdset, NULL);
    handle_io(&r_fdset, &w_fdset, &e_fdset, fds, is_finished);
  }
  return 1;
}



/**************************************************************************************
 **************************************************************************************/


int vsh_initialize(int is_interactive) {
  //debug_set_mode(1);

  vsh = vsh_new();

  vsh->fd = connect_to_daemon(0, 0, 0LL, 0);
  vsh->id = 1LL;
  vsh->is_interactive = is_interactive;

  /* setup vsh->host_vp, vsh->vp, and  */
  vsh_send(vsh->fd, -1LL, 0, JOIN);
  vsh_recv(vsh->fd);

  assert(vsh->is_initialized);
  vsh_print();

  return 1;
}


void vsh_finalize() {
  vsh_dispose(vsh);
}


/**************************************************************************************
 **************************************************************************************/


void vsh_set_pid(pid_t pid) {
  if ((vsh == NULL) || (vsh->is_initialized == 0)) {
    return;
  }

  current_task->pid = pid;
  task_list_insert(vsh->tasks, current_task);
  task_list_print(vsh->tasks);
}


void vsh_delete_task(pid_t pid) {
  task_t task;

  if ((vsh == NULL) || (vsh->is_initialized == 0)) {
    return;
  }

  if ((task = task_list_find_same_pid(vsh->tasks, pid)) != NULL)  {
    task_list_delete(vsh->tasks, task);
  }
}


/**************************************************************************************
 **************************************************************************************/


void vsh_handle_signals(int sig, pid_t pid) {
}



void vsh_get_machine_list() {
  if ((vsh == NULL) || (vsh->is_initialized == 0)) {
    return;
  }

  vsh_send(vsh->fd, vsh->host_vp, 0, GET_MACHINE_LIST);
  vsh_recv(vsh->fd);
}


/**************************************************************************************
 **************************************************************************************/


/**************************************************************************************
 **************************************************************************************/



static int bin_vpg_show(char ** argv, char * ops, int func) {
  int show_machines = 0;

  assert((argv != NULL) && (Streq(argv[0], "show")));


  if ((argv[1] == NULL) || (Streq(argv[1], "machines"))) {
    show_machines = 1;
  }


  if (show_machines) {
    ph_vp_range_cell_t p;

    printf("[machines]\n");

    for (p = vsh->machines->head; p != NULL; p = p->next) {
      char s[BUFSIZE];
      
      vp2nickname(p->self->lower, s, BUFSIZE);
      printf("%s%s", s, ((p->next != NULL) ? ", " : ""));
    }
    printf("\n");
  }
  fflush(stdout);

  return 0;
}



static int bin_vpg_kill(char ** argv, char * ops, int func) {
  return 0;
}


static int bin_vpg_jobs(char ** argv, char * ops, int func) {
  return 0;
}



int bin_vpg(char *name, char **argv, char *ops, int func) {
  if ((vsh == NULL) || (vsh->is_initialized == 0)) {
    return 0;
  }

  if ((argv != NULL) && (argv[0] != NULL)) {
    int i;
    
    for (i = 0; i < V_BUILTIN_N; i++) {
      if (Streq(v_builtins[i].name, argv[0])) {
	return (*v_builtins[i].func)(argv, ops, func);
      }
    }
  }
  
  return 0;
}

